-----------------------------------------
-- ALU
-----------------------------------------

---- Description --------------------------------------------------
-- ALU Function
---- Arithmetic
---- Compare
---- Logic
---- Shift
--
-- cf."߃Zbg߂܂傤"
-- OP is operation code (op2)
-- SRC1 := R2, SRC2 := R3 or Imm(sign extended);
-- R1 := DST;
-------------------------------------------------------------------

---- Notes --------------------------------------------------------
-- 2007.02.23
--   gݍ킹HŁB
--   ̓Iɂ́Awhen-elseɏBfunctionvSzcB
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;

entity ALU is
    port( OP : in std_logic_vector(3 downto 0);          --- opcode
          SRC1,SRC2 : in std_logic_vector(31 downto 0);  --- R2 -> SRC1, R3 -> SRC2
          DST : out std_logic_vector(31 downto 0));      --- output
end ALU;


architecture RTL of ALU is
    
    --Shift߁BABbitshiftBOɂEVtgVtg𕪂ĂB
    function shift(A: std_logic_vector(31 downto 0);B: integer;O:std_logic_vector(1 downto 0)) return std_logic_vector is
   
      variable SINTL,SINTRL,SINTRA : Integer;
      constant ALL0 : std_logic_vector(31 downto 0) := (others => '0');
      constant ALL1 : std_logic_vector(31 downto 0) := (others => '1');
      variable RET : std_logic_vector(31 downto 0);

      begin
        case O is
           when "00" => -- sl,sli
               if B >= 32 then
                   return ALL0;
               else
                   return A(32-B downto 0) & ALL0(B-1 downto 0);
               end if;
           when "01" => -- srl,srli
               if B >= 32 then
                   return ALL0;
               else
                   return ALL0(B-1 downto 0) & A(31 downto B);
               end if;
           when "10" => -- sra,srai
               if B >= 32 then
                   RET := (others => A(31));
                   return RET;
               else
                   if A(31) = '1' then
                       return ALL1(B-1 downto 0) & A(31 downto B);
                   else
                       return ALL0(B-1 downto 0) & A(31 downto B);
                   end if;
               end if;
           when others => return "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
       end case;
 end shift;
 
 function mul(A:std_logic_vector(31 downto 0);B:std_logic_vector(31 downto 0);S:std_logic) return std_logic_vector is
    variable V : std_logic_vector(63 downto 0);
 begin
     V := A * B;
     if S = '0' then
         return V(31 downto 0);
     else
        return V(63 downto 32);
      end if;
 end mul;

begin

    DST <= -- arith
            SRC1 + SRC2 when OP = "0000" else
            SRC1 - SRC2 when OP = "0001" else
            mul(SRC1,SRC2,'0') when OP = "0010" else
            mul(SRC1,SRC2,'1') when OP = "0011" else
            CONV_STD_LOGIC_VECTOR(CONV_INTEGER(SRC1) / CONV_INTEGER(SRC2),32) when OP = "0100" else
           -- compare
            (others => '1') when ((OP = "0101") and (SRC1 = SRC2)) else  -- eq,eqi
            (others => '0') when ((OP = "0101") and (SRC1 /= SRC2)) else
            (others => '1') when ((OP = "0110") and (SRC1 < SRC2)) else  -- lt,lti
            (others => '0') when ((OP = "0110") and (SRC1 >= SRC2)) else
            (others => '1') when ((OP = "0111") and (SRC1 > SRC2)) else  -- gti
            (others => '0') when ((OP = "0111") and (SRC1 <= SRC2)) else
           -- logic
            SRC1 and SRC2 when OP = "1000" else -- and,andi
            SRC1 or SRC2 when OP = "1001" else -- or,ori
            SRC1 xor SRC2 when OP = "1010" else -- xor,xori
            not SRC1 when OP = "1011" else     -- not
            -- shift
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"00") when OP = "1100" else -- sl,sli
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"01") when OP = "1101" else -- srl,srli
            shift(SRC1,CONV_INTEGER('0' & SRC2(5 downto 0)),"10") when OP = "1110" else -- sra,srai
            (others => 'X');

end RTL;
