-- FINVSQRTの繰り返し計算部分
-- X(n+1) = (3 - A*X(n)*X(n))*X(n)/2を行う
-- 

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity FINVSQRT_repeat is

	port(X      : in std_logic_vector(24 downto 0);    --MSB含む
	     A      : in std_logic_vector(24 downto 0);
	     nextX : out std_logic_vector(24 downto 0));

end FINVSQRT_repeat;

architecture STRUCTURE of FINVSQRT_repeat is
--signal AX : std_logic_vector(49 downto 0);         -- A*X(n)
--signal AXX : std_logic_vector(74 downto 0);        -- A*X(n)*X(n)
--signal AXX_buf : std_logic_vector(74 downto 0);
--signal ThreeMinusAXX : std_logic_vector(27 downto 0);
--signal nextX_buf : std_logic_vector(52 downto 0);
--
----signal Ah : std_logic_vector(12 downto 0);
----signal Al : std_logic_vector(11 downto 0);
--signal Xh : std_logic_vector(12 downto 0);
--signal Xl : std_logic_vector(11 downto 0);
----signal AXhh : std_logic_vector(25 downto 0);
----signal AXhl : std_logic_vector(24 downto 0);
-----signal AXlh : std_logic_vector(24 downto 0);
----signal AXll : std_logic_vector(23 downto 0);
----signal AXh : std_logic_vector(12 downto 0);
----signal AXl : std_logic_vector(11 downto 0);
----signal AXXhh : std_logic_vector(25 downto 0);
----signal AXXhl : std_logic_vector(24 downto 0);
----signal AXXlh : std_logic_vector(24 downto 0);
----signal AXXll : std_logic_vector(23 downto 0);
--signal nextXhh : std_logic_vector(25 downto 0);
--signal nextXhl : std_logic_vector(24 downto 0);
--signal nextXlh : std_logic_vector(24 downto 0);
--signal nextXll : std_logic_vector(23 downto 0);

signal AX : std_logic_vector(49 downto 0);         -- A*X(n)
signal AXX : std_logic_vector(74 downto 0);        -- A*X(n)*X(n)
signal AXX_buf : std_logic_vector(74 downto 0);
signal ThreeMinusAXX : std_logic_vector(27 downto 0);
signal nextX_buf : std_logic_vector(52 downto 0);

signal Ah : std_logic_vector(12 downto 0);
signal Al : std_logic_vector(11 downto 0);
signal Xh : std_logic_vector(12 downto 0);
signal Xl : std_logic_vector(11 downto 0);
signal AXhh : std_logic_vector(25 downto 0);
signal AXhl : std_logic_vector(24 downto 0);
signal AXlh : std_logic_vector(24 downto 0);
signal AXll : std_logic_vector(23 downto 0);
--signal AXh : std_logic_vector(24 downto 0);
--signal AXl : std_logic_vector(24 downto 0);
--signal AXXhh : std_logic_vector(37 downto 0);
--signal AXXhl : std_logic_vector(36 downto 0);
--signal AXXlh : std_logic_vector(37 downto 0);
--signal AXXll : std_logic_vector(36 downto 0);
signal nextXhh : std_logic_vector(26 downto 0);
signal nextXhl : std_logic_vector(26 downto 0);
signal nextXlh : std_logic_vector(25 downto 0);
signal nextXll : std_logic_vector(25 downto 0);

begin -- STRUCTURE

--　↓とりあえず動く
----Ah <= A(24 downto 12);
----Al <= A(11 downto 0);
----Xh <= X(24 downto 12);
----Xl <= X(11 downto 0);
----AXhh <= Ah * Xh;
----AXhl <= Ah * Xl;
----AXlh <= Al * Xh;
----AXll <= Al * Xl;
----AX <= (AXhh & AXll(23 downto 12)) + ("0000000000000" & AXhl) + ("0000000000000" & AXlh);
--AX <= X * A; -- A*X(n) X,Aともに最上位bitは1
--
----AXh <= AX(37 downto 25);
----AXl <= AX(24 downto 13);
----AXXhh <= AXh * Xh;
----AXXhl <= AXh * Xl;
----AXXlh <= AXl * Xh;
----AXXll <= AXl * Xl;
----AXX <= (AXXhh & AXXll(23 downto 12)) + ("0000000000000" & AXXhl) + ("0000000000000" & AXXlh);
--AXX_buf <= X * AX;
--AXX <= AXX_buf(73 downto 0) & '0' when AXX_buf(74 downto 73) = "00" else AXX_buf;
---- AXX_bufの頭は"001"か"01"か'1'（掛け算により、先頭の'1'が最大で1つ右にずれる）
--
--ThreeMinusAXX <= "1100000000000000000000000000" - ('0' & AXX(74 downto 48));--("0" & AXX(37 downto 13));
--
----nextXhh <= ThreeMinusAXX(24 downto 12) * Xh;
------nextXhl <= ThreeMinusAXX(11 downto  0) * Xh;
----nextXlh <= ThreeMinusAXX(24 downto 12) * Xl;
----nextXll <= ThreeMinusAXX(11 downto  0) * Xl;
----nextX_buf <= (nextXhh & nextXll(23 downto 12)) + ("0000000000000" & nextXhl) + ("0000000000000" & nextXlh);
--nextX_buf <= ThreeMinusAXX * X;
--
--nextX <= nextX_buf(52 downto 28) when nextX_buf(52) = '1' else
--         nextX_buf(51 downto 27);
--　↑とりあえず動く

Ah <= A(24 downto 12);
Al <= A(11 downto 0);
Xh <= X(24 downto 12);
Xl <= X(11 downto 0);
AXhh <= Ah * Xh;
AXhl <= Ah * Xl;
AXlh <= Al * Xh;
AXll <= Al * Xl;
--AX <= ((AXhh & AXll(23 downto 12)) + ("000000000000" & (('0' & AXhl) + ('0' & AXlh)))) & AXll(11 downto 0);
AX <= (AXhh & AXll) + ("000000000000" & (('0' & AXhl) + ('0' & AXlh))  & "000000000000");
--Ah <= A(24 downto 18);
--Al <= A(17 downto 0);
--Xh <= X(24 downto 18);
--Xl <= X(17 downto 0);
--AXhh <= Ah * Xh;
--AXhl <= Ah * Xl;
--AXlh <= Al * Xh;
--AXll <= Al * Xl;
--AX <= ((AXhh & AXll(35 downto 18)) + ("0000000" & (('0' & AXhl) + ('0' & AXlh)))) & AXll(17 downto 0);
--AX <= (AXhh & AXll) + ("0000000" & (('0' & AXhl) + ('0' & AXlh)) & "000000000000000000");
--AX <= X * A; -- A*X(n) X,Aともに最上位bitは1

--AXh <= AX(49 downto 25);
--AXl <= AX(24 downto 0);
--AXXhh <= AXh * Xh;
--AXXhl <= AXh * Xl;
--AXXlh <= AXl * Xh;
--AXXll <= AXl * Xl;
--AXX_buf <= (AXXhh & AXXll) + ("0000000000000" & AXXhl & "0000000000000000000000000") + ("0000000000000000000000000" & AXXlh & "000000000000");
AXX_buf <= X * AX;
AXX <= AXX_buf(73 downto 0) & '0' when AXX_buf(74 downto 73) = "00" else AXX_buf;
-- AXX_bufの頭は"001"か"01"か'1'（掛け算により、先頭の'1'が最大で1つ右にずれる）

ThreeMinusAXX <= "1100000000000000000000000000" - ('0' & AXX(74 downto 48));--("0" & AXX(37 downto 13));


--nextXhh <= ThreeMinusAXX(27 downto 18) * X(24 downto 18);
--nextXhl <= ThreeMinusAXX(17 downto  0) * X(24 downto 18);
--nextXlh <= ThreeMinusAXX(27 downto 18) * X(17 downto 0);
--nextXll <= ThreeMinusAXX(17 downto  0) * X(17 downto 0);
--nextX_buf <= (nextXhh & nextXll(35 downto 18)) + ("0000000000" & nextXhl) + ("0000000" & nextXlh);
nextXhh <= ThreeMinusAXX(27 downto 14) * Xh;
nextXhl <= ThreeMinusAXX(13 downto  0) * Xh;
nextXlh <= ThreeMinusAXX(27 downto 14) * Xl;
nextXll <= ThreeMinusAXX(13 downto  0) * Xl;
nextX_buf <= (nextXhh & nextXll) + (("00000000000000" & nextXhl & "000000000000") + ("0000000000000" & nextXlh & "00000000000000"));

--nextX_buf <= ThreeMinusAXX * X;

nextX <= nextX_buf(52 downto 28) when nextX_buf(52) = '1' else
         nextX_buf(51 downto 27);


--↓ボツ
--Ah <= A(24 downto 12);
--Al <= A(11 downto 0);
--Xh <= X(24 downto 12);
--Xl <= X(11 downto 0);
--AXhh <= Ah * Xh;
--AXhl <= Ah * Xl;
--AXlh <= Al * Xh;
--AXll <= Al * Xl;
--AX <= (AXhh & AXll(23 downto 12)) + ("0000000000000" & AXhl) + ("0000000000000" & AXlh);
----AX <= X * A; -- A*X(n) X,Aともに最上位bitは1
--
--AXh <= AX(37 downto 25);
--AXl <= AX(24 downto 13);
--AXXhh <= AXh * Xh;
--AXXhl <= AXh * Xl;
--AXXlh <= AXl * Xh;
--AXXll <= AXl * Xl;
--AXX <= (AXXhh & AXXll(23 downto 12)) + ("0000000000000" & AXXhl) + ("0000000000000" & AXXlh);
----AXX <= X * AX;
--
--ThreeMinusAXX <= "11000000000000000000000000" - ("00" & AXX(37 downto 14));--("0" & AXX(37 downto 13));
--
--nextXhh <= ThreeMinusAXX(24 downto 12) * Xh;
--nextXhl <= ThreeMinusAXX(11 downto  0) * Xh;
--nextXlh <= ThreeMinusAXX(24 downto 12) * Xl;
--nextXll <= ThreeMinusAXX(11 downto  0) * Xl;
--nextX_buf <= (nextXhh & nextXll(23 downto 12)) + ("0000000000000" & nextXhl) + ("0000000000000" & nextXlh);
----nextX_buf <= ThreeMinusAXX(24 downto 0) * X;
--
--nextX <= nextX_buf(37 downto 13) when nextX_buf(37) = '1' else
--         nextX_buf(36 downto 12);
--　↑ボツ          
end STRUCTURE;
