-- FADD_S
-- ソースオペランドの符合は一致しているものとする

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity FADD_S is

	port(SRC1,SRC2 : in std_logic_vector(31 downto 0);
	     OUTPUT    : out std_logic_vector(31 downto 0);
	     NaNout    : out std_logic;
	     ZeroOut   : out std_logic);

end FADD_S;

architecture STRUCTURE of FADD_S is
component RightShifter24
	port(Data    : in std_logic_vector(23 downto 0);
	     nShift  : in std_logic_vector(7 downto 0);
	     Shifted : out std_logic_vector(25 downto 0));
end component;

signal NaN : std_logic;                       -- Not a Number
signal Inf : std_logic;                       -- Infinity
signal Zero : std_logic;                      -- Zero

signal S : std_logic; 
signal W : std_logic_vector(31 downto 0);     -- W数
signal L : std_logic_vector(31 downto 0);     -- L数
signal Data : std_logic_vector(23 downto 0);  -- input to RightShifter
signal nShift : std_logic_vector(7 downto 0); -- input to RightShifter
signal EXP : std_logic_vector(7 downto 0);
signal EXP_buf : std_logic_vector(7 downto 0);
signal L_MANT_Shifted : std_logic_vector(25 downto 0); -- L数の仮数部を右シフトしたもの
signal MANT : std_logic_vector(22 downto 0);           -- MANT_buf2のMSBを左端にそろえたもの
signal MANT_buf : std_logic_vector(25 downto 0);
signal MANT_buf2 : std_logic_vector(25 downto 0);      -- MANT_bufを丸め

signal O :std_logic;
signal Os : std_logic_vector(8 downto 0);
begin -- STRUCTURER
   NaN <= '1' when (SRC1(30 downto 23) = 255 and SRC1(22 downto 0) /= 0) 
                or (SRC2(30 downto 23) = 255 and SRC2(22 downto 0) /= 0) else
          '0';
   ---NaN <= '1' when (SRC1(30 downto 23) = 255 or SRC2(30 downto 23) = 255) else
   ---       '0';
   Inf <= '1' when SRC1(30 downto 0) = "1111111100000000000000000000000" or
                   SRC2(30 downto 0) = "1111111100000000000000000000000" or
                   EXP = 255 else
         '0';
   Zero <= '1' when (SRC1(30 downto 0) = 0 and SRC2(30 downto 0) = 0) or
               EXP = 0 else
           '0';
    
   S <= SRC1(31);
   L <= SRC1 when SRC1(30 downto 23) < SRC2(30 downto 23) else SRC2;
   W <= SRC2 when SRC1(30 downto 23) < SRC2(30 downto 23) else SRC1;
   EXP_buf <= W(30 downto 23); 
   Data <= '1' & L(22 downto 0);                -- RightShifter24の入力
   nShift <= W(30 downto 23) - L(30 downto 23); -- RightShifter24の入力
   
   --RightShifter24_1 : RightShifter24 port map (Data,nShift,L_MANT_Shifted); 
   L_MANT_Shifted <= shr(Data&"00",nShift);
   
   MANT_buf <= ("01" & W(22 downto 0) & '0') + ('0' & L_MANT_Shifted(25 downto 1));
   EXP <= EXP_buf + 1 when (MANT_buf(25) = '1' or MANT_buf(24 downto 0) = "1111111111111111111111111") else
          EXP_buf;
   --MANT_buf2 <= (MANT_buf(26 downto 2) + 1) & MANT_buf(1 downto 0) when (MANT_buf(26) = '0' and MANT_buf(1) = '1') else
   --             (MANT_buf(26 downto 3) + 1) & MANT_buf(2 downto 0) when (MANT_buf(26) = '1' and MANT_buf(2) = '1') else
   MANT_buf2 <= MANT_buf + 2 when (MANT_buf(25) = '0' and MANT_buf(0) = '1') else
                MANT_buf + 4 when (MANT_buf(25) = '1' and MANT_buf(1) = '1') else
                MANT_buf; -- 丸め
   MANT <= MANT_buf2(24 downto 2) when MANT_buf2(25) = '1' else
           MANT_buf2(23 downto 1);
   
   NaNout <= NaN;
   ZeroOut <= Zero and (not Inf);
   
   O <= NaN or (not Zero);
   Os <= (others => O); 
   OUTPUT <= --"1111111100000000000000000000000" when Inf = '1' and NaN = '0' else
             --Os & "0000000000000000000000"   when (NaN or Zero) = '1' else
             ----"01111111110000000000000000000000" when NaN = '1' else
             ----"00000000000000000000000000000000" when Zero = '1' else
             --EXP & MANT;
             
             S & "1111111100000000000000000000000"   when Inf = '1' and NaN = '0' else
             '0' & Os & "0000000000000000000000" when (NaN or Zero) = '1' else
             --"01111111110000000000000000000000" when NaN = '1' else
             --"00000000000000000000000000000000" when Zero = '1' else
             S & EXP & MANT;
end STRUCTURE;
