-- FDIV
-- 引き算を繰り返して4bitずつ計算していく
-- 

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity FDIV is

	port(SRC1,SRC2 : in std_logic_vector(31 downto 0);
	     OUTPUT    : out std_logic_vector(31 downto 0));

end FDIV;

architecture STRUCTURE of FDIV is
component Div_4bit_Sub is
	port(IN_A,IN_B : in std_logic_vector(23 downto 0);
	     IN_C      : in std_logic_vector(27 downto 0);
	     RESTin   : in std_logic;
	     REST     : out std_logic;
	     OUT_A    : out std_logic_vector(23 downto 0);
	     OUT_C    : out std_logic_vector(27 downto 0));
end component;

signal S : std_logic;
signal Zero1 : std_logic;                  -- '1' when SRC1 = 0
signal Zero2 : std_logic;
signal NaN : std_logic;                    -- SRC2が0のときはNansを返す
signal Inf1 : std_logic;                   -- '1' when SRC1 is Infinity
signal Inf2 : std_logic;

signal A : std_logic_vector(23 downto 0); -- MANT of SRC1
signal B : std_logic_vector(23 downto 0); -- MANT of SRC2
signal C : std_logic_vector(27 downto 0) := "0000000000000000000000000000"; -- A/Bを入れる
signal WinA : std_logic;
signal EXP : std_logic_vector(9 downto 0);
signal C1to2 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal C2to3 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal C3to4 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal C4to5 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal C5to6 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal C6to7 : std_logic_vector(27 downto 0);-- := "0000000000000000000000000000";
signal A1to2 : std_logic_vector(23 downto 0);
signal A2to3 : std_logic_vector(23 downto 0);
signal A3to4 : std_logic_vector(23 downto 0);
signal A4to5 : std_logic_vector(23 downto 0);
signal A5to6 : std_logic_vector(23 downto 0);
signal A6to7 : std_logic_vector(23 downto 0);
signal REST : std_logic := '0';
signal REST1to2 : std_logic;
signal REST2to3 : std_logic;
signal REST3to4 : std_logic;
signal REST4to5 : std_logic;
signal REST5to6 : std_logic;
signal REST6to7 : std_logic;
signal REST7to8 : std_logic;

signal OUT_A : std_logic_vector(23 downto 0);
signal OUT_C : std_logic_vector(27 downto 0);

signal EXP0 : std_logic;       -- '1' when EXP =< 0
signal EXP255 : std_logic;     -- '1' when EXP => 255

begin -- STRUCTURE
   Zero1 <= '1' when SRC1(30 downto 0) = "0000000000000000000000000000000" else
            '0'; 
   Zero2 <= '1' when SRC2(30 downto 0) = "0000000000000000000000000000000" else
            '0'; 
   NaN <= '1' when (SRC1(30 downto 23) = 255 and SRC1(22 downto 0) /= 0) or 
                   (SRC2(30 downto 23) = 255 and SRC2(22 downto 0) /= 0) else
          '0';
   Inf1 <= '1' when SRC1(30 downto 0) = "1111111100000000000000000000000" else
           '0';
   Inf2 <= '1' when SRC2(30 downto 0) = "1111111100000000000000000000000" else
           '0';
          
   S <= SRC1(31) xor SRC2(31);
   A <= '1' & SRC1(22 downto 0);
   B <= '1' & SRC2(22 downto 0);
   WinA <= '1' when A >= B else
           '0';
   
   EXP <= ("00" & SRC1(30 downto 23)) - ("00" & SRC2(30 downto 23)) + 127 when WinA = '1' else
          ("00" & SRC1(30 downto 23)) - ("00" & SRC2(30 downto 23)) + 126;
   
   EXP0 <= '1' when EXP(7 downto 0) = 0 or EXP(9) = '1' else '0';
   EXP255 <= '1' when EXP(7 downto 0) = 255 or EXP(8) = '1' else '0';
   
   --Div_4bit_Sub_0 : Div_4bit_Sub port map (A,B,C,OUT_A,OUT_C); -- for test
   Div_4bit_Sub_1 : Div_4bit_Sub port map (A    ,B,C    ,REST    ,REST1to2,A1to2,C1to2);
   Div_4bit_Sub_2 : Div_4bit_Sub port map (A1to2,B,C1to2,REST1to2,REST2to3,A2to3,C2to3);
   Div_4bit_Sub_3 : Div_4bit_Sub port map (A2to3,B,C2to3,REST2to3,REST3to4,A3to4,C3to4);
   Div_4bit_Sub_4 : Div_4bit_Sub port map (A3to4,B,C3to4,REST3to4,REST4to5,A4to5,C4to5);
   Div_4bit_Sub_5 : Div_4bit_Sub port map (A4to5,B,C4to5,REST4to5,REST5to6,A5to6,C5to6);
   Div_4bit_Sub_6 : Div_4bit_Sub port map (A5to6,B,C5to6,REST5to6,REST6to7,A6to7,C6to7);
   Div_4bit_Sub_7 : Div_4bit_Sub port map (A6to7,B,C6to7,REST6to7,REST7to8,OUT_A,OUT_C);
   
   OUTPUT <= "01111111110000000000000000000000"    when NaN = '1' or (Inf1 = '1' and Inf2 = '1') else--or (Zero1 = '1' and Zero2 = '1') else -- Nans
             "00000000000000000000000000000000"    when Zero1 = '1' or Inf2 = '1' or EXP0 = '1' else      -- Zero
             SRC1(31) & "1111111100000000000000000000000" when Zero2 = '1' else
             S & "1111111100000000000000000000000" when Inf1 = '1' or EXP255 = '1' else
             --OUT_C & "0000";                               -- for test
             S & EXP(7 downto 0) & OUT_C(26 downto 4) when WinA = '1' else
             S & EXP(7 downto 0) & OUT_C(25 downto 3); -- WinA = '0'
                        
end STRUCTURE;

