-- FMUL

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity FMUL is

	port(SRC1,SRC2 : in std_logic_vector(31 downto 0);
	     OUTPUT    : out std_logic_vector(31 downto 0));

end FMUL;

architecture STRUCTURE of FMUL is
signal NaN : std_logic;                       -- Not a Number
signal Inf : std_logic;                       -- Infinity
signal Zero : std_logic;                      -- Zero

signal S : std_logic;
signal EXP_buf : std_logic_vector(9 downto 0); -- SRC1+SRC2-127
signal EXP : std_logic_vector(9 downto 0);
signal EXP_sub1 : std_logic_vector(9 downto 0); -- EXP - 1
signal EXP_add1 : std_logic_vector(9 downto 0); -- EXP + 1
signal MANT_buf : std_logic_vector(26 downto 0); -- SRC1,SRC2の仮数部をそのまま掛けたもの
--signal MANT_buf : std_logic_vector(47 downto 0);
--signal MANT_buf2 : std_logic_vector(26 downto 0); -- MANT_bufの後ろ22bitを丸めたもの
signal MANT_buf3 : std_logic_vector(23 downto 0); -- LeftShift後
signal MANT_buf4 : std_logic_vector(22 downto 0); -- 丸め後

signal SMANT1_high : std_logic_vector(11 downto 0); --24bit×24bitの分割に使用
signal SMANT1_low  : std_logic_vector(11 downto 0);
signal SMANT2_high : std_logic_vector(11 downto 0);
signal SMANT2_low  : std_logic_vector(11 downto 0);
signal SMANT_hh : std_logic_vector(23 downto 0); -- SMANT1_high × SMANT2_high
signal SMANT_hl : std_logic_vector(23 downto 0);
signal SMANT_lh : std_logic_vector(23 downto 0);
signal SMANT_ll : std_logic_vector(23 downto 0);

--signal EXP_buf2 : std_logic_vector(9 downto 0);
signal O :std_logic;
signal Os : std_logic_vector(8 downto 0);

begin -- STRUCTURE
   NaN <= '1' when (SRC1(30 downto 23) = 255 and SRC1(22 downto 0) /= 0) 
                or (SRC2(30 downto 23) = 255 and SRC2(22 downto 0) /= 0) else
          '0';
   Inf <= '1' when SRC1(30 downto 0) = "1111111100000000000000000000000" or
                   SRC2(30 downto 0) = "1111111100000000000000000000000" or
                   (EXP_add1(8) = '1' and EXP_sub1(9) = '0') else
          '0';
   Zero <= '1' when SRC1(30 downto 0) = 0 or SRC2(30 downto 0) = 0 or
                    EXP_sub1(9) = '1'  else
           '0';
   
   S <= SRC1(31) xor SRC2(31); -- 符号
   EXP_buf <= ("00" & SRC1(30 downto 23)) + ("00" & SRC2(30 downto 23)) - "0001111111"; -- 仮の指数部
   --EXP_buf <= SRC1(30 downto 23) + SRC2(30 downto 23) - "01111111"; -- 仮の指数部
   EXP <= EXP_buf + 1
      when (MANT_buf(26) = '1' or MANT_buf(25 downto 0) = "01111111111111111111111111") else
          EXP_buf;
   --EXP <= "00" & EXP_buf(7 downto 0) + 1 
   --   when (MANT_buf2(26) = '1' or MANT_buf2(26 downto 1) = "01111111111111111111111111") else
   --       "00" & EXP_buf(7 downto 0);
   EXP_sub1 <= EXP - 1; -- exp    is negative when EXP_sub1(9) = '1'
   EXP_add1 <= EXP + 1; -- output is infinity when EXP_add1(8) = '1'
   
   SMANT1_high <= '1' & SRC1(22 downto 12);
   SMANT1_low  <= SRC1(11 downto 0);
   SMANT2_high <= '1' & SRC2(22 downto 12);
   SMANT2_low  <= SRC2(11 downto 0);
   SMANT_hh <= (SMANT1_high * SMANT2_high);
   SMANT_hl <= SMANT1_high * SMANT2_low;
   SMANT_lh <= SMANT1_low * SMANT2_high;
   SMANT_ll <= (SMANT1_low * SMANT2_low);
   MANT_buf <= (SMANT_hh & SMANT_ll(23 downto 21)) + ("00000000000" & (('0' & SMANT_hl(23 downto 11)) + ('0' & SMANT_lh(23 downto 11))));
   --MANT_buf <= ('1' & SRC1(22 downto 0)) * ('1' & SRC2(22 downto 0)); --24bit×24bitを上記のように分割
   
--   MANT_buf2 <= MANT_buf(47 downto 22) & '1' when MANT_buf(21 downto 0) /= "0000000000000000000000"
--            else MANT_buf(47 downto 22) & '0';   
--   MANT_buf3 <= MANT_buf2(26 downto 1) when MANT_buf2(26) = '1' else -- 桁揃え
--                MANT_buf2(25 downto 0);-- when MANT_buf2(26) = '0';
--   MANT_buf4 <= MANT_buf3(25 downto 2) + 1 when ((MANT_buf3(1) and (MANT_buf3(2) or MANT_buf3(0))) = '1') else
--                MANT_buf3(25 downto 2); -- 丸め

   MANT_buf3 <= MANT_buf(25 downto 2) when MANT_buf(26) = '1' else -- 桁揃え
                MANT_buf(24 downto 1);-- when MANT_buf(26) = '0';
   MANT_buf4 <= MANT_buf3(23 downto 1) + 1 when (MANT_buf3(0) = '1') else
                MANT_buf3(23 downto 1); -- 丸め

   O <= (not Zero) or NaN or (Inf and Zero);
   Os <= (others => O); --Nansの場合とZeroの場合をまとめる
   OUTPUT <= --"01111111110000000000000000000000"    when NaN = '1' or (Inf = '1' and Zero = '1') else
             '0' & Os & "0000000000000000000000" when (Zero or NaN or (Inf and Zero)) = '1' else
             S & "1111111100000000000000000000000" when Inf = '1' else
             --"00000000000000000000000000000000"    when Zero = '1' else
             S & EXP(7 downto 0) & MANT_buf4;
end STRUCTURE;
