----------------------------------------
-- Instruction Fetch & Decode & Read Phase
----------------------------------------

-------------------------------------------------------------------------------
-- Notes
-- 2007.01.27
---- sram̃f[^̃^C~OɊւāAiĵŁACB
---- pipeline register𓱓BKv̂́AB
--  2007.01.26
---- ƂA!
---- l͂̕ʓ|ɂȂăXg[B
---- l܂ŏBꉞňʂ芮Ȃ̂ȁB
---- IԂBֈڍs`XȂB
---- ̍HvȂǁAIԂւ̈ڍs܂B
-- 2007.02.11
---- PC̑͊OɏoǂBƎvďo܂B
---- ̂Ƃ̓RgAEgB
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity IFDR is
    port(CLK : in std_logic;     -- system clock
        RESET : in std_logic;    -- reset
        -- Mr.SRAM&EXW
        PC : out std_logic_vector(19 downto 0);  -- instruction address(PC) ->sram & EXW
        -- SRAM
        INST : in std_logic_vector(31 downto 0); -- instruction
        -- EXW
        EXPC : in std_logic_vector(19 downto 0);   -- External PC
        BUSY : in std_logic;                       -- EXW busy signal
        VALID : out std_logic;                    -- FDR valid
        OP1 : out std_logic_vector(1 downto 0);   -- opcode1
        OP2 : out std_logic_vector(3 downto 0);   -- opcode2
        REGAD : out std_logic_vector(6 downto 0);  -- desg reg addr/imm7,disp7
        IMMSEL : out std_logic;                    -- if IMM then 1 else 0
        IMM : out std_logic_vector(31 downto 0);   -- imm
        -- register file
        REG_ADS : out std_logic_vector(6 downto 0); -- R2 addr
        REG_ADT : out std_logic_vector(6 downto 0) -- R3 addr
        );
end IFDR;


architecture ARC of IFDR is

    signal STATE : std_logic_vector(3 downto 0);
    signal NFPC,NPC : std_logic_vector(19 downto 0);
    signal PCSEL : std_logic;                      -- 0->NPC 1->EXPC
    signal PCSELB : std_logic;
    
    -- pipeline buffer
    signal OP1B : std_logic_vector(1 downto 0);
    signal OP2B : std_logic_vector(3 downto 0);
    signal REGADB : std_logic_vector(6 downto 0);  -- desg reg addr/imm7,disp7
    signal IMMSELB : std_logic;                    -- if IMM then 1 else 0
    signal IMMB : std_logic_vector(31 downto 0);   -- imm
    signal REG_ADSB,REG_ADTB : std_logic_vector(6 downto 0);

begin

    OP1 <= OP1B;
    OP2 <= OP2B;
    REGAD <= REGADB;
    IMMSEL <= IMMSELB;
    IMM <= IMMB;
    REG_ADS <= REG_ADSB;
    REG_ADT <= REG_ADTB;

    PC <= NFPC;

    process(CLK,RESET)
    begin
        if RESET = '1' then
            STATE <= "0100";
            NPC <= (others => '0');
            --NFPC <= (others => '0');
            PCSEL <= '0';
            PCSELB <= '0';
            VALID <= '0';
            --PC <= (others => '0');
        elsif CLK'event and CLK = '1' then
            case STATE is
                -- init
                when "0100" =>
                    NFPC <= NPC;
                    NPC <= NPC + 1;
                    STATE <= STATE + 1;
                when "0101" =>
                    STATE <= STATE + 1;
                when "0110" =>
                    STATE <= "0000";
                -- cycle
                when "0000" =>
                    -- next fetch
                    if PCSEL = '1' then
                        NFPC <= EXPC;
                        NPC <= EXPC + 1;
                    else
                        NFPC <= NPC;
                        NPC <= NPC + 1;
                    end if;
                    -- current decode
                    OP1B <= INST(31 downto 30);
                    OP2B <= INST(22 downto 19);
                    REGADB <=INST(29 downto 23);
                    IMMSELB <= INST(18);
                    -- next PC
                    if INST(31 downto 30) = "11" then
                        PCSEL <= '1';
                    else
                        PCSEL <= '0';
                    end if;
                    -- IMM11/IMM21/disp19 & Regaddr exception
                    case INST(31 downto 30) is
                        when "00" =>
                            if INST(10) = '1' and INST(22) = '0' then
                                IMMB <= "111111111111111111111" & INST(10 downto 0);
                            else
                                IMMB <= "000000000000000000000" & INST(10 downto 0);
                            end if;
                            REG_ADTB <= INST(10 downto 4);  --r3
                        when "10" =>
                            if INST(22 downto 21) = "00" then
                                IMMB <= "00000000000" & INST(20 downto 0);
                            else
                                IMMB <= "000000000000000000000" & INST(10 downto 0);
                            end if;
                            if (INST(18) xor INST(19)) = '1' then
                                REG_ADTB <= INST(10 downto 4);  --r3
                            else
                                REG_ADTB <= INST(29 downto 23); --r1
                            end if;
                        when "11" =>
                            IMMB <= "000000000000" & INST(18) & INST(18 downto 0);
                            REG_ADTB <= INST(29 downto 23); --r1
                        when others =>
                            if INST(10) = '1' then
                                IMMB <= "111111111111111111111" & INST(10 downto 0);
                            else
                                IMMB <= "000000000000000000000" & INST(10 downto 0);
                            end if;
                            REG_ADTB <= INST(10 downto 4);  --r3
                    end case;
                    REG_ADSB <= INST(17 downto 11); --r2
                    --VALID <= '1';
                    if (INST(31 downto 30) = "11") and (INST(22 downto 19) = "0111") then
                        STATE <= "1111";
                    else
                        STATE <= STATE + 1;
                    end if;
                when "0001" =>
                    if BUSY = '0' then
                       VALID <= '1';
                       STATE <= STATE + 1;
                    else
                       STATE <= "0001";
                    end if;
                    --if BUSY = '0' then
                    --    VALID <= '0';
                    --    STATE <= STATE + 1;
                        --PCSELB <= PCSEL;
                    -- next fetch
                    --   if PCSELB = '1' then
                    --       NFPC <= EXPC;
                    --       NPC <= EXPC + 1;
                    --   else
                    --       NFPC <= NPC;
                    --       NPC <= NPC + 1;
                    --   end if;                        
                    --else
                    --    STATE <= "0001";
                    --end if;
                when "0010" =>
                    --if BUSY = '0' then
                        VALID <= '0';
                        STATE <= STATE + 1;
                        --PCSELB <= PCSEL;
                    -- next fetch
                    --   if PCSELB = '1' then
                    --       NFPC <= EXPC;
                    --       NPC <= EXPC + 1;
                    --   else
                    --       NFPC <= NPC;
                    --       NPC <= NPC + 1;
                    --   end if;                        
                    --else
                    --    VALID <= '1';
                    --    STATE <= "0010";
                    --    OP1B <= OP1B;
                    --    OP2B <= OP2B;
                    --    REGADB <= REGADB;
                    --    IMMSELB <= IMMSELB;
                    --    IMMB <= IMMB;
                    --    REG_ADSB <= REG_ADSB;
                    --    REG_ADTB <= REG_ADTB;
                    --end if;
                    --STATE <= "0000";
                    --STATE <= STATE + 1;
                when "0011" =>
                    STATE <= "0000";
                --when "1000" =>
                --when "1000" =>
                --    VALID <= '1';
                --    if BUSY = '0' then
                        --- 
                --    else
                --       OP1B <= OP1B;
                --      OP2B <= OP2B;
                --      REGADB <= REGADB;
                --      IMMSELB <= IMMSELB;
                --      IMMB <= IMMB;
                --      REG_ADSB <= REG_ADSB;
                --     REG_ADTB <= REG_ADTB;
                --       STATE <= "1000";
                --   end if;
                   
                when "1111" =>  -- END
                    STATE <= "1111";
                when others =>
                    STATE <= "0000";
            end case;
        end if;
    end process;
end ARC;

